# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
from pyomo.core.expr import inequality, RangedExpression, EqualityExpression
from pyomo.kernel import pprint
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.constraint import (
    IConstraint,
    constraint,
    linear_constraint,
    constraint_dict,
    constraint_tuple,
    constraint_list,
)
from pyomo.core.kernel.variable import variable
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.expression import expression, data_expression
from pyomo.core.kernel.block import block


class Test_constraint(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        v = variable()
        c = constraint((1, v**2, 2))
        pprint(c)
        b = block()
        b.c = c
        pprint(c)
        pprint(b)
        m = block()
        m.b = b
        pprint(c)
        pprint(b)
        pprint(m)

    def test_ctype(self):
        c = constraint()
        self.assertIs(c.ctype, IConstraint)
        self.assertIs(type(c), constraint)
        self.assertIs(type(c)._ctype, IConstraint)

    def test_pickle(self):
        c = constraint()
        self.assertIs(c.lb, None)
        self.assertIs(c.body, None)
        self.assertIs(c.ub, None)
        self.assertEqual(c.parent, None)
        cup = pickle.loads(pickle.dumps(c))
        self.assertEqual(cup.lb, None)
        self.assertEqual(cup.body, None)
        self.assertEqual(cup.ub, None)
        self.assertEqual(cup.parent, None)
        b = block()
        b.c = c
        self.assertIs(c.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        cup = bup.c
        self.assertEqual(cup.lb, None)
        self.assertEqual(cup.body, None)
        self.assertEqual(cup.ub, None)
        self.assertIs(cup.parent, bup)

    def test_init(self):
        c = constraint()
        self.assertTrue(c.parent is None)
        self.assertEqual(c.ctype, IConstraint)
        self.assertIs(c.body, None)
        self.assertIs(c.lb, None)
        self.assertIs(c.ub, None)
        self.assertEqual(c.equality, False)
        with self.assertRaises(ValueError):
            self.assertEqual(c(), None)
        with self.assertRaises(ValueError):
            self.assertEqual(c(exception=True), None)
        self.assertEqual(c(exception=False), None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

    def test_has_lb_ub(self):
        c = constraint()
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        c.lb = float('-inf')
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), False)
        self.assertIs(c.ub, None)

        c.ub = float('inf')
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        c.lb = 0
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(type(c.lb), int)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        c.ub = 0
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(type(c.lb), int)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)
        self.assertEqual(type(c.ub), int)

        c.lb = float('inf')
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(type(c.lb), float)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)
        self.assertEqual(type(c.ub), int)

        c.ub = float('-inf')
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(type(c.lb), float)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(type(c.ub), float)

        c.rhs = float('inf')
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(type(c.lb), float)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        c.rhs = float('-inf')
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(type(c.ub), float)

        c.equality = False
        pL = parameter()
        c.lb = pL
        self.assertIs(c.lower, pL)
        pU = parameter()
        c.ub = pU
        self.assertIs(c.upper, pU)

        with self.assertRaises(ValueError):
            self.assertEqual(c.has_lb(), False)
        self.assertIs(c.lower, pL)
        with self.assertRaises(ValueError):
            self.assertEqual(c.has_ub(), False)
        self.assertIs(c.upper, pU)

        pL.value = float('-inf')
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        with self.assertRaises(ValueError):
            self.assertEqual(c.has_ub(), False)
        self.assertIs(c.upper, pU)

        pU.value = float('inf')
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        pL.value = 0
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        pU.value = 0
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)

        pL.value = float('inf')
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)

        pU.value = float('-inf')
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, float('-inf'))

        pL.value = float('inf')
        c.rhs = pL
        self.assertEqual(c.has_lb(), True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.has_ub(), False)
        self.assertEqual(c.ub, None)

        pL.value = float('-inf')
        c.rhs = pL
        self.assertEqual(c.has_lb(), False)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, float('-inf'))

    def test_bounds_getter_setter(self):
        c = constraint()
        self.assertEqual(c.bounds, (None, None))
        self.assertEqual(c.lb, None)
        self.assertEqual(c.ub, None)

        c.bounds = (1, 2)
        self.assertEqual(c.bounds, (1, 2))
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 2)

        c.rhs = 3
        self.assertEqual(c.bounds, (3, 3))
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)
        with self.assertRaises(ValueError):
            c.bounds = (3, 3)
        self.assertEqual(c.bounds, (3, 3))
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)
        with self.assertRaises(ValueError):
            c.bounds = (2, 2)
        self.assertEqual(c.bounds, (3, 3))
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)
        with self.assertRaises(ValueError):
            c.bounds = (1, 2)
        self.assertEqual(c.bounds, (3, 3))
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)

    def test_init_nonexpr(self):
        v = variable()
        c = constraint(lb=0, body=v, ub=1)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, v)
        self.assertEqual(c.ub, 1)
        with self.assertRaises(ValueError):
            constraint(lb=0, expr=v <= 1)
        with self.assertRaises(ValueError):
            constraint(body=v, expr=v <= 1)
        with self.assertRaises(ValueError):
            constraint(ub=1, expr=v <= 1)
        with self.assertRaises(ValueError):
            constraint(rhs=1, expr=v <= 1)
        c = constraint(expr=v <= 1)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, v)
        self.assertEqual(c.ub, 1)

        with self.assertRaises(ValueError):
            constraint(rhs=1, lb=1)
        with self.assertRaises(ValueError):
            constraint(rhs=1, ub=1)
        c = constraint(rhs=1)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertIs(c.body, None)

    def test_type(self):
        c = constraint()
        self.assertTrue(isinstance(c, ICategorizedObject))
        self.assertTrue(isinstance(c, IConstraint))

    def test_active(self):
        c = constraint()
        self.assertEqual(c.active, True)
        c.deactivate()
        self.assertEqual(c.active, False)
        c.activate()
        self.assertEqual(c.active, True)

        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        b.c = c
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, False)
        c.deactivate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        b.activate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, True)
        b.activate(shallow=False)
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        b.deactivate(shallow=False)
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)

    def test_equality(self):
        v = variable()
        c = constraint(v == 1)
        self.assertTrue(c.body is not None)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.equality, True)

        c = constraint(1 == v)
        self.assertTrue(c.body is not None)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.equality, True)

        c = constraint(v - 1 == 0)
        self.assertTrue(c.body is not None)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 0)
        self.assertEqual(c.rhs, 0)
        self.assertEqual(c.equality, True)

        c = constraint(0 == v - 1)
        self.assertTrue(c.body is not None)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 0)
        self.assertEqual(c.rhs, 0)
        self.assertEqual(c.equality, True)

        c = constraint(rhs=1)
        self.assertIs(c.body, None)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.equality, True)

        # can not set when equality is True
        with self.assertRaises(ValueError):
            c.lb = 2
        # can not set when equality is True
        with self.assertRaises(ValueError):
            c.ub = 2

        c.equality = False

        # can not get when equality is False
        with self.assertRaises(ValueError):
            c.rhs

        self.assertIs(c.body, None)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)

        # can not set to True, must set rhs to a value
        with self.assertRaises(ValueError):
            c.equality = True

        c.rhs = 3
        self.assertIs(c.body, None)
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)
        self.assertEqual(c.equality, True)

        with self.assertRaises(TypeError):
            c.rhs = 'a'

        with self.assertRaises(ValueError):
            c.rhs = None

    def test_nondata_bounds(self):
        c = constraint()
        e = expression()

        eL = expression()
        eU = expression()
        with self.assertRaises(ValueError):
            c.expr = (eL, e, eU)
        e.expr = 1.0
        eL.expr = 1.0
        eU.expr = 1.0
        with self.assertRaises(ValueError):
            c.expr = (eL, e, eU)
        with self.assertRaises(TypeError):
            c.lb = eL
        with self.assertRaises(TypeError):
            c.ub = eU

        vL = variable()
        vU = variable()
        with self.assertRaises(ValueError):
            c.expr = (vL, e, vU)
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU

        e.expr = 1.0
        vL.value = 1.0
        vU.value = 1.0
        with self.assertRaises(ValueError):
            c.expr = (vL, e, vU)
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU
        with self.assertRaises(TypeError):
            c.rhs = vL

        # the fixed status of a variable
        # does not change this restriction
        vL.fixed = True
        vU.fixed = True
        with self.assertRaises(ValueError):
            c.expr = (vL, e, vU)
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU
        with self.assertRaises(TypeError):
            c.rhs = vL

        vL.value = 1.0
        vU.value = 1.0
        with self.assertRaises(ValueError):
            c.expr = (vL, 0.0, vU)
        c.body = -2.0
        c.lb = 1.0
        c.ub = 1.0
        self.assertEqual(c.slack, -3.0)
        self.assertEqual(c.lslack, -3.0)
        self.assertEqual(c.uslack, 3.0)

        with self.assertRaises(TypeError):
            c.lb = 'a'
        with self.assertRaises(TypeError):
            c.ub = 'a'
        self.assertEqual(c.lb, 1.0)
        self.assertEqual(c.ub, 1.0)

        vL.value = 2
        vU.value = 1
        c.expr = vL <= vU
        self.assertEqual(c.lb, None)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)
        c.expr = vU >= vL
        self.assertEqual(c.lb, None)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)
        c.expr = vU <= vL
        self.assertEqual(c.lb, None)
        self.assertEqual(c.body(), -1)
        self.assertEqual(c.ub, 0)
        c.expr = vL >= vU
        self.assertEqual(c.lb, None)
        self.assertEqual(c.body(), -1)
        self.assertEqual(c.ub, 0)

    def test_fixed_variable_stays_in_body(self):
        c = constraint()
        x = variable(value=0.5)
        c.expr = (0, x, 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 0.5)
        self.assertEqual(c.ub, 1)
        x.value = 2
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 2)
        self.assertEqual(c.ub, 1)

        # ensure the variable is not moved into the upper or
        # lower bound expression (this used to be a bug)
        x.fix(0.5)
        c.expr = (0, x, 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 0.5)
        self.assertEqual(c.ub, 1)
        x.value = 2
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 2)
        self.assertEqual(c.ub, 1)

        x.free()
        x.value = 1
        c.expr = 0 == x
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)
        c.expr = x == 0
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)

        # ensure the variable is not moved into the upper or
        # lower bound expression (this used to be a bug)
        x.fix()
        c.expr = 0 == x
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)
        c.expr = x == 0
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)

        # ensure the variable is not moved into the upper or
        # lower bound expression (this used to be a bug)
        x.free()
        c.expr = 0 == x
        x.fix()
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)
        x.free()
        c.expr = x == 0
        x.fix()
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.ub, 0)

    def test_data_bounds(self):
        c = constraint()
        e = expression(expr=1.0)

        pL = parameter()
        pU = parameter()
        c.expr = (pL, e, pU)
        self.assertIs(c.body, e)
        self.assertIs(c.lower, pL)
        self.assertIs(c.upper, pU)
        e.expr = None
        self.assertIs(c.body, e)
        self.assertIs(c.lower, pL)
        self.assertIs(c.upper, pU)
        c.expr = (pL, e, pU)
        self.assertIs(c.body, e)
        self.assertIs(c.lower, pL)
        self.assertIs(c.upper, pU)

        e.expr = 1.0
        eL = data_expression()
        eU = data_expression()
        c.expr = (eL, e, eU)
        self.assertIs(c.body, e)
        self.assertIs(c.lower, eL)
        self.assertIs(c.upper, eU)
        e.expr = None
        self.assertIs(c.body, e)
        self.assertIs(c.lower, eL)
        self.assertIs(c.upper, eU)
        c.expr = (eL, e, eU)
        self.assertIs(c.body, e)
        self.assertIs(c.lower, eL)
        self.assertIs(c.upper, eU)

    # make sure we can use a mutable param that
    # has not been given a value in the upper bound
    # of an inequality constraint
    def test_mutable_novalue_param_lower_bound(self):
        x = variable()
        p = parameter()
        p.value = None

        c = constraint(expr=0 <= x - p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p <= x)
        self.assertIs(c.lower, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p <= x + 1)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p + 1 <= x)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p + 1) ** 2 <= x)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p, x, p + 1))
        self.assertEqual(c.equality, False)

        c = constraint(expr=x - p >= 0)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x >= p)
        self.assertIs(c.lower, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x + 1 >= p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x >= p + 1)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x >= (p + 1) ** 2)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p, x, None))
        self.assertIs(c.lower, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p, x + 1, None))
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p + 1, x, None))
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p, x, 1))
        self.assertEqual(c.equality, False)

    # make sure we can use a mutable param that
    # has not been given a value in the lower bound
    # of an inequality constraint
    def test_mutable_novalue_param_upper_bound(self):
        x = variable()
        p = parameter()
        p.value = None

        c = constraint(expr=x - p <= 0)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x <= p)
        self.assertIs(c.upper, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x + 1 <= p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x <= p + 1)
        self.assertEqual(c.equality, False)

        c = constraint(expr=x <= (p + 1) ** 2)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p + 1, x, p))
        self.assertEqual(c.equality, False)

        c = constraint(expr=0 >= x - p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p >= x)
        self.assertIs(c.upper, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p >= x + 1)
        self.assertEqual(c.equality, False)

        c = constraint(expr=p + 1 >= x)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(p + 1) ** 2 >= x)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(None, x, p))
        self.assertIs(c.upper, p)
        self.assertEqual(c.equality, False)

        c = constraint(expr=(None, x + 1, p))
        self.assertEqual(c.equality, False)

        c = constraint(expr=(None, x, p + 1))
        self.assertEqual(c.equality, False)

        c = constraint(expr=(1, x, p))
        self.assertEqual(c.equality, False)

    # make sure we can use a mutable param that
    # has not been given a value in the rhs of
    # of an equality constraint
    def test_mutable_novalue_param_equality(self):
        x = variable()
        p = parameter()
        p.value = None

        c = constraint(expr=x - p == 0)
        self.assertEqual(c.equality, True)

        c = constraint(expr=x == p)
        self.assertIs(c.upper, p)
        self.assertEqual(c.equality, True)

        c = constraint(expr=x + 1 == p)
        self.assertEqual(c.equality, True)

        c = constraint(expr=x + 1 == (p + 1) ** 2)
        self.assertEqual(c.equality, True)

        c = constraint(expr=x == p + 1)
        self.assertEqual(c.equality, True)

        c = constraint(expr=(x, p))
        self.assertIs(c.upper, p)
        self.assertIs(c.lower, p)
        self.assertIs(c.rhs, p)
        self.assertIs(c.body, x)
        self.assertEqual(c.equality, True)

        c = constraint(expr=(p, x))
        self.assertIs(c.upper, p)
        self.assertIs(c.lower, p)
        self.assertIs(c.rhs, p)
        self.assertIs(c.body, x)
        self.assertEqual(c.equality, True)

        c = constraint(expr=EqualityExpression((p, x)))
        self.assertIs(c.upper, p)
        self.assertIs(c.lower, p)
        self.assertIs(c.rhs, p)
        self.assertIs(c.body, x)
        self.assertEqual(c.equality, True)

        c = constraint(expr=EqualityExpression((x, p)))
        self.assertIs(c.upper, p)
        self.assertIs(c.lower, p)
        self.assertIs(c.rhs, p)
        self.assertIs(c.body, x)
        self.assertEqual(c.equality, True)

    def test_tuple_construct_equality(self):
        x = variable()
        c = constraint((0.0, x))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(type(c.lb), float)
        self.assertIs(c.body, x)
        self.assertEqual(c.ub, 0)
        self.assertEqual(type(c.ub), float)

        c = constraint((x, 0))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(type(c.lb), int)
        self.assertIs(c.body, x)
        self.assertEqual(c.ub, 0)
        self.assertEqual(type(c.ub), int)

    def test_tuple_construct_inf_equality(self):
        x = variable()
        c = constraint((x, float('inf')))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertEqual(type(c.rhs), float)
        self.assertIs(c.body, x)
        c = constraint((float('inf'), x))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertEqual(type(c.rhs), float)
        self.assertIs(c.body, x)

    def test_tuple_construct_1sided_inequality(self):
        y = variable()
        c = constraint((None, y, 1))
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, 1)

        c = constraint((0, y, None))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

    def test_tuple_construct_1sided_inf_inequality(self):
        y = variable()
        c = constraint((float('-inf'), y, 1))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, 1)
        self.assertEqual(type(c.ub), int)

        c = constraint((0, y, float('inf')))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, 0)
        self.assertEqual(type(c.lb), int)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, None)

    def test_tuple_construct_unbounded_inequality(self):
        y = variable()
        c = constraint((None, y, None))
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

        c = constraint((float('-inf'), y, float('inf')))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, None)

    def test_tuple_construct_invalid_1sided_inequality(self):
        x = variable()
        y = variable()
        z = variable()
        with self.assertRaises(ValueError):
            constraint((x, y, None))

        with self.assertRaises(ValueError):
            constraint((None, y, z))

    def test_tuple_construct_2sided_inequality(self):
        y = variable()
        c = constraint((0, y, 1))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, 1)

    def test_construct_invalid_2sided_inequality(self):
        x = variable()
        y = variable()
        z = variable()
        with self.assertRaises(ValueError):
            constraint((x, y, 1))

        with self.assertRaises(ValueError):
            constraint((0, y, z))

    def test_tuple_construct_invalid_2sided_inequality(self):
        x = variable()
        y = variable()
        z = variable()
        with self.assertRaises(ValueError):
            constraint(RangedExpression((x, y, 1), (False, False)))

        with self.assertRaises(ValueError):
            constraint(RangedExpression((0, y, z), (False, False)))

    def test_expr_construct_equality(self):
        x = variable(value=1)
        y = variable(value=1)
        c = constraint(0.0 == x)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, x)
        self.assertEqual(c.ub, 0)

        c = constraint(x == 0.0)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, x)
        self.assertEqual(c.ub, 0)

        c = constraint(x == y)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertTrue(c.body is not None)
        self.assertEqual(c(), 0)
        self.assertEqual(c.body(), 0)
        self.assertEqual(c.ub, 0)

        c = constraint()
        c.expr = x == float('inf')
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, x)
        c.expr = float('inf') == x
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, x)

    def test_strict_inequality_failure(self):
        x = variable()
        y = variable()
        c = constraint()
        with self.assertRaises(ValueError):
            c.expr = x < 0
        with self.assertRaises(ValueError):
            c.expr = inequality(body=x, upper=0, strict=True)
        c.expr = x <= 0
        c.expr = inequality(body=x, upper=0, strict=False)
        with self.assertRaises(ValueError):
            c.expr = x > 0
        with self.assertRaises(ValueError):
            c.expr = inequality(body=x, lower=0, strict=True)
        c.expr = x >= 0
        c.expr = inequality(body=x, lower=0, strict=False)
        with self.assertRaises(ValueError):
            c.expr = x < y
        with self.assertRaises(ValueError):
            c.expr = inequality(body=x, upper=y, strict=True)
        c.expr = x <= y
        c.expr = inequality(body=x, upper=y, strict=False)
        with self.assertRaises(ValueError):
            c.expr = x > y
        with self.assertRaises(ValueError):
            c.expr = inequality(body=x, lower=y, strict=True)
        c.expr = x >= y
        c.expr = inequality(body=x, lower=y, strict=False)
        with self.assertRaises(ValueError):
            c.expr = RangedExpression((0, x, 1), (True, True))
        with self.assertRaises(ValueError):
            c.expr = RangedExpression((0, x, 1), (False, True))
        with self.assertRaises(ValueError):
            c.expr = RangedExpression((0, x, 1), (True, False))
        c.expr = RangedExpression((0, x, 1), (False, False))

    def test_expr_construct_inf_equality(self):
        x = variable()
        c = constraint(x == float('inf'))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, x)
        c = constraint(float('inf') == x)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, x)

    def test_expr_construct_1sided_inequality(self):
        y = variable()
        c = constraint(y <= 1)
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, 1)

        c = constraint(0 <= y)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

        c = constraint(y >= 1)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, 1)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

        c = constraint(0 >= y)
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, 0)

    def test_expr_construct_unbounded_inequality(self):
        y = variable()
        c = constraint(y <= float('inf'))
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, None)

        c = constraint(float('-inf') <= y)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, None)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

        c = constraint(y >= float('-inf'))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, None)
        self.assertIs(c.body, y)
        self.assertIs(c.ub, None)

        c = constraint(float('inf') >= y)
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, y)
        self.assertEqual(c.ub, None)

    def test_expr_construct_unbounded_inequality(self):
        y = variable()
        c = constraint(y <= float('-inf'))
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertIs(c.body, y)
        c = constraint(float('inf') <= y)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, float('inf'))
        self.assertIs(c.ub, None)
        self.assertIs(c.body, y)
        c = constraint(y >= float('inf'))
        self.assertEqual(c.equality, False)
        self.assertEqual(c.lb, float('inf'))
        self.assertIs(c.ub, None)
        self.assertIs(c.body, y)
        c = constraint(float('-inf') >= y)
        self.assertEqual(c.equality, False)
        self.assertIs(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertIs(c.body, y)

    def test_expr_invalid_double_sided_inequality(self):
        x = variable()
        y = variable()
        c = constraint()
        c.expr = (0, x - y, 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)
        with self.assertRaises(ValueError):
            c.expr = (x, x - y, 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)
        with self.assertRaises(ValueError):
            c.expr = (0, x - y, y)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)
        with self.assertRaises(ValueError):
            c.expr = (1, x - y, x)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)
        with self.assertRaises(ValueError):
            c.expr = (y, x - y, 0)

    def test_equality_infinite(self):
        c = constraint()
        v = variable()
        c.expr = v == 1
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertIs(c.body, v)
        c.expr = v == float('inf')
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, v)
        c.expr = (v, float('inf'))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, v)
        c.expr = float('inf') == v
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, v)
        c.expr = (float('inf'), v)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, float('inf'))
        self.assertEqual(c.ub, None)
        self.assertEqual(c.rhs, float('inf'))
        self.assertIs(c.body, v)
        c.expr = v == float('-inf')
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(c.rhs, float('-inf'))
        self.assertIs(c.body, v)
        c.expr = (v, float('-inf'))
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(c.rhs, float('-inf'))
        self.assertIs(c.body, v)
        c.expr = float('-inf') == v
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(c.rhs, float('-inf'))
        self.assertIs(c.body, v)
        c.expr = (float('-inf'), v)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.ub, float('-inf'))
        self.assertEqual(c.rhs, float('-inf'))
        self.assertIs(c.body, v)

    def test_equality_nonnumeric(self):
        c = constraint()
        v = variable()
        c.expr = v == 1
        with self.assertRaises(TypeError):
            c.expr = (v, 'x')
        with self.assertRaises(TypeError):
            c.expr = ('x', v)

    def test_slack_methods(self):
        x = variable(value=2)
        L = 1
        U = 5

        # equality
        cE = constraint(rhs=L, body=x)
        x.value = 4
        self.assertEqual(cE.body(), 4)
        self.assertEqual(cE.slack, -3)
        self.assertEqual(cE.lslack, 3)
        self.assertEqual(cE.uslack, -3)
        x.value = 6
        self.assertEqual(cE.body(), 6)
        self.assertEqual(cE.slack, -5)
        self.assertEqual(cE.lslack, 5)
        self.assertEqual(cE.uslack, -5)
        x.value = 0
        self.assertEqual(cE.body(), 0)
        self.assertEqual(cE.slack, -1)
        self.assertEqual(cE.lslack, -1)
        self.assertEqual(cE.uslack, 1)
        x.value = None
        with self.assertRaises(ValueError):
            cE.body()
        self.assertEqual(cE.body(exception=False), None)
        self.assertEqual(cE.slack, None)
        self.assertEqual(cE.lslack, None)
        self.assertEqual(cE.uslack, None)

        # equality
        cE = constraint(rhs=U, body=x)
        x.value = 4
        self.assertEqual(cE.body(), 4)
        self.assertEqual(cE.slack, -1)
        self.assertEqual(cE.lslack, -1)
        self.assertEqual(cE.uslack, 1)
        x.value = 6
        self.assertEqual(cE.body(), 6)
        self.assertEqual(cE.slack, -1)
        self.assertEqual(cE.lslack, 1)
        self.assertEqual(cE.uslack, -1)
        x.value = 0
        self.assertEqual(cE.body(), 0)
        self.assertEqual(cE.slack, -5)
        self.assertEqual(cE.lslack, -5)
        self.assertEqual(cE.uslack, 5)
        x.value = None
        with self.assertRaises(ValueError):
            cE.body()
        self.assertEqual(cE.body(exception=False), None)
        self.assertEqual(cE.slack, None)
        self.assertEqual(cE.lslack, None)
        self.assertEqual(cE.uslack, None)

        # lower finite
        cL = constraint(lb=L, body=x)
        x.value = 4
        self.assertEqual(cL.body(), 4)
        self.assertEqual(cL.slack, 3)
        self.assertEqual(cL.lslack, 3)
        self.assertEqual(cL.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cL.body(), 6)
        self.assertEqual(cL.slack, 5)
        self.assertEqual(cL.lslack, 5)
        self.assertEqual(cL.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cL.body(), 0)
        self.assertEqual(cL.slack, -1)
        self.assertEqual(cL.lslack, -1)
        self.assertEqual(cL.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cL.body()
        self.assertEqual(cL.body(exception=False), None)
        self.assertEqual(cL.slack, None)
        self.assertEqual(cL.lslack, None)
        self.assertEqual(cL.uslack, None)

        # lower unbounded
        cL = constraint(lb=float('-inf'), body=x)
        x.value = 4
        self.assertEqual(cL.body(), 4)
        self.assertEqual(cL.slack, float('inf'))
        self.assertEqual(cL.lslack, float('inf'))
        self.assertEqual(cL.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cL.body(), 6)
        self.assertEqual(cL.slack, float('inf'))
        self.assertEqual(cL.lslack, float('inf'))
        self.assertEqual(cL.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cL.body(), 0)
        self.assertEqual(cL.slack, float('inf'))
        self.assertEqual(cL.lslack, float('inf'))
        self.assertEqual(cL.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cL.body()
        self.assertEqual(cL.body(exception=False), None)
        self.assertEqual(cL.slack, None)
        self.assertEqual(cL.lslack, None)
        self.assertEqual(cL.uslack, None)

        # upper finite
        cU = constraint(body=x, ub=U)
        x.value = 4
        self.assertEqual(cU.body(), 4)
        self.assertEqual(cU.slack, 1)
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, 1)
        x.value = 6
        self.assertEqual(cU.body(), 6)
        self.assertEqual(cU.slack, -1)
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, -1)
        x.value = 0
        self.assertEqual(cU.body(), 0)
        self.assertEqual(cU.slack, 5)
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, 5)
        x.value = None
        with self.assertRaises(ValueError):
            cU.body()
        self.assertEqual(cU.body(exception=False), None)
        self.assertEqual(cU.slack, None)
        self.assertEqual(cU.lslack, None)
        self.assertEqual(cU.uslack, None)

        # upper unbounded
        cU = constraint(body=x, ub=float('inf'))
        x.value = 4
        self.assertEqual(cU.body(), 4)
        self.assertEqual(cU.slack, float('inf'))
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cU.body(), 6)
        self.assertEqual(cU.slack, float('inf'))
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cU.body(), 0)
        self.assertEqual(cU.slack, float('inf'))
        self.assertEqual(cU.lslack, float('inf'))
        self.assertEqual(cU.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cU.body()
        self.assertEqual(cU.body(exception=False), None)
        self.assertEqual(cU.slack, None)
        self.assertEqual(cU.lslack, None)
        self.assertEqual(cU.uslack, None)

        # range finite
        cR = constraint(lb=L, body=x, ub=U)
        x.value = 4
        self.assertEqual(cR.body(), 4)
        self.assertEqual(cR.slack, 1)
        self.assertEqual(cR.lslack, 3)
        self.assertEqual(cR.uslack, 1)
        x.value = 6
        self.assertEqual(cR.body(), 6)
        self.assertEqual(cR.slack, -1)
        self.assertEqual(cR.lslack, 5)
        self.assertEqual(cR.uslack, -1)
        x.value = 0
        self.assertEqual(cR.body(), 0)
        self.assertEqual(cR.slack, -1)
        self.assertEqual(cR.lslack, -1)
        self.assertEqual(cR.uslack, 5)
        x.value = None
        with self.assertRaises(ValueError):
            cR.body()
        self.assertEqual(cR.body(exception=False), None)
        self.assertEqual(cR.slack, None)
        self.assertEqual(cR.lslack, None)
        self.assertEqual(cR.uslack, None)

        # range unbounded (None)
        cR = constraint(body=x)
        x.value = 4
        self.assertEqual(cR.body(), 4)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cR.body(), 6)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cR.body(), 0)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cR.body()
        self.assertEqual(cR.body(exception=False), None)
        self.assertEqual(cR.slack, None)
        self.assertEqual(cR.lslack, None)
        self.assertEqual(cR.uslack, None)

        # range unbounded
        cR = constraint(body=x, lb=float('-inf'), ub=float('inf'))
        x.value = 4
        self.assertEqual(cR.body(), 4)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cR.body(), 6)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cR.body(), 0)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cR.body()
        self.assertEqual(cR.body(exception=False), None)
        self.assertEqual(cR.slack, None)
        self.assertEqual(cR.lslack, None)
        self.assertEqual(cR.uslack, None)

        # range finite (parameter)
        cR = constraint(body=x, lb=parameter(L), ub=parameter(U))
        x.value = 4
        self.assertEqual(cR.body(), 4)
        self.assertEqual(cR.slack, 1)
        self.assertEqual(cR.lslack, 3)
        self.assertEqual(cR.uslack, 1)
        x.value = 6
        self.assertEqual(cR.body(), 6)
        self.assertEqual(cR.slack, -1)
        self.assertEqual(cR.lslack, 5)
        self.assertEqual(cR.uslack, -1)
        x.value = 0
        self.assertEqual(cR.body(), 0)
        self.assertEqual(cR.slack, -1)
        self.assertEqual(cR.lslack, -1)
        self.assertEqual(cR.uslack, 5)
        x.value = None
        with self.assertRaises(ValueError):
            cR.body()
        self.assertEqual(cR.body(exception=False), None)
        self.assertEqual(cR.slack, None)
        self.assertEqual(cR.lslack, None)
        self.assertEqual(cR.uslack, None)

        # range unbounded (parameter)
        cR = constraint(body=x, lb=parameter(float('-inf')), ub=parameter(float('inf')))
        x.value = 4
        self.assertEqual(cR.body(), 4)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 6
        self.assertEqual(cR.body(), 6)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = 0
        self.assertEqual(cR.body(), 0)
        self.assertEqual(cR.slack, float('inf'))
        self.assertEqual(cR.lslack, float('inf'))
        self.assertEqual(cR.uslack, float('inf'))
        x.value = None
        with self.assertRaises(ValueError):
            cR.body()
        self.assertEqual(cR.body(exception=False), None)
        self.assertEqual(cR.slack, None)
        self.assertEqual(cR.lslack, None)
        self.assertEqual(cR.uslack, None)

    def test_expr(self):
        x = variable(value=1.0)
        c = constraint()
        c.expr = (0, x, 2)
        self.assertEqual(c(), 1)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.ub, 2)
        self.assertEqual(c.equality, False)

        c.expr = (-2, x, 0)
        self.assertEqual(c(), 1)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c.lb, -2)
        self.assertEqual(c.ub, 0)
        self.assertEqual(c.equality, False)

    def test_expr_getter(self):
        c = constraint()
        self.assertIs(c.expr, None)

        v = variable()

        c.expr = 0 <= v
        self.assertIsNot(c.expr, None)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, v)
        self.assertIs(c.ub, None)
        self.assertEqual(c.equality, False)

        c.expr = v <= 1
        self.assertIsNot(c.expr, None)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, v)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)

        c.expr = (0, v, 1)
        self.assertIsNot(c.expr, None)
        self.assertEqual(c.lb, 0)
        self.assertIs(c.body, v)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)

        c.expr = v == 1
        self.assertIsNot(c.expr, None)
        self.assertEqual(c.lb, 1)
        self.assertIs(c.body, v)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, True)

        c.expr = None
        self.assertIs(c.expr, None)
        self.assertIs(c.lb, None)
        self.assertIs(c.body, None)
        self.assertIs(c.ub, None)
        self.assertEqual(c.equality, False)

    def test_expr_wrong_type(self):
        c = constraint()
        with self.assertRaises(ValueError):
            c.expr = 2
        with self.assertRaises(ValueError):
            c.expr = True

    def test_tuple_constraint_create(self):
        x = variable()
        y = variable()
        z = variable()
        c = constraint((0.0, x))
        with self.assertRaises(ValueError):
            constraint((y, x, z))
        with self.assertRaises(ValueError):
            constraint((0, x, z))
        with self.assertRaises(ValueError):
            constraint((y, x, 0))
        with self.assertRaises(ValueError):
            constraint((x, 0, 0, 0))

        c = constraint((x, y))
        self.assertEqual(c.upper, 0)
        self.assertEqual(c.lower, 0)
        self.assertTrue(c.body is not None)

    def test_expression_constructor_coverage(self):
        x = variable()
        y = variable()
        z = variable()
        L = parameter(value=0)
        U = parameter(value=1)

        expr = U >= x
        expr = expr >= L
        c = constraint(expr)

        expr = x <= z
        expr = expr >= y
        with self.assertRaises(ValueError):
            constraint(expr)

        expr = x >= z
        expr = y >= expr
        with self.assertRaises(ValueError):
            constraint(expr)

        expr = y <= x
        expr = y >= expr
        with self.assertRaises(ValueError):
            constraint(expr)

        L.value = 0
        c = constraint(x >= L)

        U.value = 0
        c = constraint(U >= x)

        L.value = 0
        U.value = 1
        expr = U <= x
        expr = expr <= L
        c = constraint(expr)

        expr = x >= z
        expr = expr <= y
        with self.assertRaises(ValueError):
            constraint(expr)

        expr = x <= z
        expr = y <= expr
        with self.assertRaises(ValueError):
            constraint(expr)

        expr = y >= x
        expr = y <= expr
        with self.assertRaises(ValueError):
            constraint(expr)

        L.value = 0
        expr = x <= L
        c = constraint(expr)

        U.value = 0
        expr = U <= x
        c = constraint(expr)

        x = variable()
        with self.assertRaises(ValueError):
            constraint(x + x)


class Test_linear_constraint(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        v = variable()
        c = linear_constraint(lb=1, terms=[(v, 1)], ub=1)
        pprint(c)
        b = block()
        b.c = c
        pprint(c)
        pprint(b)
        m = block()
        m.b = b
        pprint(c)
        pprint(b)
        pprint(m)

    def test_ctype(self):
        c = linear_constraint([], [])
        self.assertIs(c.ctype, IConstraint)
        self.assertIs(type(c), linear_constraint)
        self.assertIs(type(c)._ctype, IConstraint)

    def test_pickle(self):
        c = linear_constraint([], [])
        self.assertIs(c.lb, None)
        self.assertEqual(c.body, 0)
        self.assertIs(c.ub, None)
        self.assertEqual(c.parent, None)
        cup = pickle.loads(pickle.dumps(c))
        self.assertEqual(cup.lb, None)
        self.assertEqual(cup.body, 0)
        self.assertEqual(cup.ub, None)
        self.assertEqual(cup.parent, None)
        b = block()
        b.c = c
        self.assertIs(c.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        cup = bup.c
        self.assertEqual(cup.lb, None)
        self.assertEqual(cup.body, 0)
        self.assertEqual(cup.ub, None)
        self.assertIs(cup.parent, bup)

    def test_init(self):
        c = linear_constraint([], [])
        self.assertTrue(c.parent is None)
        self.assertEqual(c.ctype, IConstraint)
        self.assertEqual(c.body, 0)
        self.assertIs(c.lb, None)
        self.assertIs(c.ub, None)
        self.assertEqual(c.equality, False)
        self.assertEqual(c(), 0)
        self.assertEqual(c.slack, float('inf'))
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(c.uslack, float('inf'))

    def test_init_nonexpr(self):
        v = variable(value=3)

        c = linear_constraint()
        self.assertEqual(len(list(c.terms)), 0)
        self.assertEqual(c.lb, None)
        self.assertEqual(c.body, 0)
        self.assertEqual(c.ub, None)

        c = linear_constraint([v], [1], lb=0, ub=1)
        self.assertEqual(len(list(c.terms)), 1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 3)
        self.assertEqual(c(), 3)
        self.assertEqual(c.ub, 1)
        # can't use both terms and variables
        with self.assertRaises(ValueError):
            linear_constraint(terms=(), variables=())
        # can't use both terms and coefficients
        with self.assertRaises(ValueError):
            linear_constraint(terms=(), coefficients=())
        # can't use both all three
        with self.assertRaises(ValueError):
            linear_constraint(terms=(), variables=(), coefficients=())
        # can't use only variables
        with self.assertRaises(ValueError):
            linear_constraint(variables=[v])
        # can't use only coefficients
        with self.assertRaises(ValueError):
            linear_constraint(coefficients=[1])
        # can't use both lb and rhs
        with self.assertRaises(ValueError):
            linear_constraint([v], [1], lb=0, rhs=0)
        # can't use both ub and rhs
        with self.assertRaises(ValueError):
            linear_constraint([v], [1], ub=0, rhs=0)

        c = linear_constraint([v], [1], rhs=1)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body(), 3)
        self.assertEqual(c(), 3)

        c = linear_constraint([], [], rhs=1)
        c.terms = ((v, 1),)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body(), 3)
        self.assertEqual(c(), 3)

    def test_init_terms(self):
        v = variable(value=3)
        c = linear_constraint([], [], rhs=1)
        c.terms = ((v, 2),)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body(), 6)
        self.assertEqual(c(), 6)

        c = linear_constraint(terms=[(v, 2)], rhs=1)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body(), 6)
        self.assertEqual(c(), 6)

        terms = [(v, 2)]
        c = linear_constraint(terms=iter(terms), rhs=1)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body(), 6)
        self.assertEqual(c(), 6)

        c.terms = ()
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.body, 0)
        self.assertEqual(c(), 0)
        self.assertEqual(tuple(c.terms), ())

    def test_type(self):
        c = linear_constraint([], [])
        self.assertTrue(isinstance(c, ICategorizedObject))
        self.assertTrue(isinstance(c, IConstraint))

    def test_active(self):
        c = linear_constraint([], [])
        self.assertEqual(c.active, True)
        c.deactivate()
        self.assertEqual(c.active, False)
        c.activate()
        self.assertEqual(c.active, True)

        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        b.c = c
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, False)
        c.deactivate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        b.activate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, True)
        b.activate(shallow=False)
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        b.deactivate(shallow=False)
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)

    def test_equality(self):
        v = variable()
        c = linear_constraint([v], [1], rhs=1)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.equality, True)

        c = linear_constraint([], [], rhs=1)
        self.assertEqual(c.body, 0)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.rhs, 1)
        self.assertEqual(c.equality, True)

        # can not set when equality is True
        with self.assertRaises(ValueError):
            c.lb = 2
        # can not set when equality is True
        with self.assertRaises(ValueError):
            c.ub = 2

        c.equality = False

        # can not get when equality is False
        with self.assertRaises(ValueError):
            c.rhs

        self.assertEqual(c.body, 0)
        self.assertEqual(c.lb, 1)
        self.assertEqual(c.ub, 1)
        self.assertEqual(c.equality, False)

        # can not set to True, must set rhs to a value
        with self.assertRaises(ValueError):
            c.equality = True

        c.rhs = 3
        self.assertEqual(c.body, 0)
        self.assertEqual(c.lb, 3)
        self.assertEqual(c.ub, 3)
        self.assertEqual(c.rhs, 3)
        self.assertEqual(c.equality, True)

    def test_nondata_bounds(self):
        c = linear_constraint([], [])

        eL = expression()
        eU = expression()
        with self.assertRaises(TypeError):
            c.rhs = eL
        with self.assertRaises(TypeError):
            c.lb = eL
        with self.assertRaises(TypeError):
            c.ub = eU

        vL = variable()
        vU = variable()
        with self.assertRaises(TypeError):
            c.rhs = vL
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU

        vL.value = 1.0
        vU.value = 1.0
        with self.assertRaises(TypeError):
            c.rhs = vL
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU

        # the fixed status of a variable
        # does not change this restriction
        vL.fixed = True
        vU.fixed = True
        with self.assertRaises(TypeError):
            c.rhs = vL
        with self.assertRaises(TypeError):
            c.lb = vL
        with self.assertRaises(TypeError):
            c.ub = vU

        vL.value = -1
        vU.value = -1
        c = linear_constraint([vL, vU], [1, 1], lb=1.0, ub=1.0)
        self.assertEqual(c(), -2.0)
        self.assertEqual(c.slack, -3.0)
        self.assertEqual(c.lslack, -3.0)
        self.assertEqual(c.uslack, 3.0)

    def test_fixed_variable_stays_in_body(self):
        x = variable(value=0.5)
        c = linear_constraint([x], [1], lb=0, ub=1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 0.5)
        self.assertEqual(c(), 0.5)
        self.assertEqual(c.ub, 1)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], x)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)
        x.value = 2
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 2)
        self.assertEqual(c(), 2)
        self.assertEqual(c.ub, 1)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], x)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

        x.fix(0.5)
        c = linear_constraint([x], [2], lb=0, ub=1)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c(), 1)
        self.assertEqual(c.ub, 1)
        repn = c.canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 1)
        x.value = 2
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 4)
        self.assertEqual(c(), 4)
        self.assertEqual(c.ub, 1)
        repn = c.canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 4)

        x.free()
        x.value = 1
        c = linear_constraint([x], [1], rhs=0)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c(), 1)
        self.assertEqual(c.ub, 0)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], x)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

        x.fix()
        c = linear_constraint([x], [1], rhs=0)
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c(), 1)
        self.assertEqual(c.ub, 0)
        repn = c.canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 1)

        x.free()
        c = linear_constraint([x], [1], rhs=0)
        x.fix()
        self.assertEqual(c.equality, True)
        self.assertEqual(c.lb, 0)
        self.assertEqual(c.body(), 1)
        self.assertEqual(c(), 1)
        self.assertEqual(c.ub, 0)
        repn = c.canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 1)

    def test_data_bounds(self):
        c = linear_constraint([], [])
        e = expression(expr=1.0)

        c.lb = 1.0
        c.ub = 2.0
        pL = parameter()
        pU = parameter()
        c.lb = pL
        self.assertIs(c.lower, pL)
        c.ub = pU
        self.assertIs(c.upper, pU)

        e.expr = 1.0
        eL = data_expression()
        eU = data_expression()
        c.lb = eL
        self.assertIs(c.lower, eL)
        c.ub = eU
        self.assertIs(c.upper, eU)

    def test_call(self):
        c = linear_constraint([], [])
        self.assertEqual(c(), 0)

        v = variable()
        c = linear_constraint([v], [2])
        with self.assertRaises(ValueError):
            c()
        with self.assertRaises(ValueError):
            c(exception=True)
        self.assertEqual(c(exception=False), None)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (2,))
        self.assertEqual(repn.constant, 0)

        v.value = 2
        self.assertEqual(c(), 4)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (2,))
        self.assertEqual(repn.constant, 0)

        v.value = None
        e = expression(v)
        c = linear_constraint([e], [1])
        with self.assertRaises(ValueError):
            c()
        with self.assertRaises(ValueError):
            c(exception=True)
        self.assertEqual(c(exception=False), None)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

        v.value = 2
        self.assertEqual(c(), 2)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

    def test_canonical_form(self):
        v = variable()
        e = expression()
        p = parameter(value=1)

        c = linear_constraint()
        self.assertEqual(c._linear_canonical_form, True)

        #
        # compute_values = True
        #

        c.terms = [(v, p)]
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

        v.fix(2)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 0)
        self.assertEqual(len(repn.linear_coefs), 0)
        self.assertEqual(repn.constant, 2)

        v.free()
        e.expr = v
        c.terms = [(e, p)]
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(repn.linear_coefs, (1,))
        self.assertEqual(repn.constant, 0)

        v.fix(2)
        repn = c.canonical_form()
        self.assertEqual(len(repn.linear_vars), 0)
        self.assertEqual(len(repn.linear_coefs), 0)
        self.assertEqual(repn.constant, 2)

        #
        # compute_values = False
        #

        v.free()
        c.terms = [(v, p)]
        repn = c.canonical_form(compute_values=False)
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(len(repn.linear_coefs), 1)
        self.assertIs(repn.linear_coefs[0], p)
        self.assertEqual(repn.linear_coefs[0](), 1)
        self.assertEqual(repn.constant, 0)

        v.fix(2)
        repn = c.canonical_form(compute_values=False)
        self.assertEqual(len(repn.linear_vars), 0)
        self.assertEqual(len(repn.linear_coefs), 0)
        self.assertEqual(repn.constant(), 2)

        v.free()
        e.expr = v
        c.terms = [(e, p)]
        repn = c.canonical_form(compute_values=False)
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], v)
        self.assertEqual(len(repn.linear_coefs), 1)
        self.assertIs(repn.linear_coefs[0], p)
        self.assertEqual(repn.linear_coefs[0](), 1)
        self.assertEqual(repn.constant, 0)

        v.fix(2)
        repn = c.canonical_form(compute_values=False)
        self.assertEqual(len(repn.linear_vars), 0)
        self.assertEqual(len(repn.linear_coefs), 0)
        self.assertEqual(repn.constant(), 2)


class Test_constraint_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = constraint_dict
    _ctype_factory = lambda self: constraint()


class Test_constraint_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = constraint_tuple
    _ctype_factory = lambda self: constraint()


class Test_constraint_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = constraint_list
    _ctype_factory = lambda self: constraint()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import logging
import weakref

from pyomo.common.gc_manager import PauseGC
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.core.base.set_types import Any
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import value
from pyomo.core.expr.numeric_expr import LinearExpression
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.constraint import IndexedConstraint, ConstraintData
from pyomo.repn.standard_repn import StandardRepn

from collections.abc import Mapping

logger = logging.getLogger('pyomo.core')


class _MatrixConstraintData(ConstraintData):
    """
    This class defines the data for a single linear constraint
        derived from a canonical form Ax=b constraint.

    Constructor arguments:
        index           The index of this component within the container.
        component       The Constraint object that owns this data.

    Public class attributes:
        active          A boolean that is true if this constraint is
                            active in the model.
        body            The Pyomo expression for this constraint
        lower           The Pyomo expression for the lower bound
        upper           The Pyomo expression for the upper bound
        equality        A boolean that indicates whether this is an
                            equality constraint
        strict_lower    A boolean that indicates whether this
                            constraint uses a strict lower bound
        strict_upper    A boolean that indicates whether this
                            constraint uses a strict upper bound

    Private class attributes:
        _component      The objective component.
        _active         A boolean that indicates whether this data is active
        _index          The row index into the main coefficient matrix
    """

    __slots__ = ()

    # the super secret flag that makes the writers
    # handle _MatrixConstraintData objects more efficiently
    _linear_canonical_form = True

    #
    # Define methods that writers expect when the
    # _linear_canonical_form flag is True
    #

    def canonical_form(self, compute_values=True):
        """Build a canonical representation of the body of
        this constraints"""
        comp = self.parent_component()
        index = self._index
        data = comp._A_data
        indices = comp._A_indices
        indptr = comp._A_indptr
        x = comp._x

        variables = []
        coefficients = []
        constant = 0
        for p in range(indptr[index], indptr[index + 1]):
            v = x[indices[p]]
            c = data[p]
            if not v.fixed:
                variables.append(v)
                if compute_values:
                    coefficients.append(value(c))
                else:
                    coefficients.append(c)
            else:
                if compute_values:
                    constant += value(c) * v.value
                else:
                    constant += c * v
        repn = StandardRepn()
        repn.linear_vars = tuple(variables)
        repn.linear_coefs = tuple(coefficients)
        repn.constant = constant
        return repn

    def __init__(self, index, component_ref):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - ConstraintData,
        #   - ActiveComponentData
        #   - ComponentData
        self._component = component_ref
        self._active = True

        # row index into the sparse matrix stored on the parent
        assert index >= 0
        self._index = index

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(_MatrixConstraintData, self).__getstate__()
        return result

    # Since this class requires no special processing of the state
    # dictionary, it does not need to implement __setstate__()

    #
    # Override the default interface methods to
    # avoid generating the body expression where
    # possible
    #

    def __call__(self, exception=NOTSET):
        """Compute the value of the body of this constraint."""
        exception = _type_check_exception_arg(self, exception)
        comp = self.parent_component()
        index = self._index
        data = comp._A_data
        indices = comp._A_indices
        indptr = comp._A_indptr
        x = comp._x
        ptrs = range(indptr[index], indptr[index + 1])
        try:
            return sum(x[indices[p]].value * data[p] for p in ptrs)
        except (ValueError, TypeError):
            if exception:
                raise
            return None

    def has_lb(self):
        """Returns :const:`False` when the lower bound is
        :const:`None` or negative infinity"""
        lb = self.lower
        return (lb is not None) and (lb != float('-inf'))

    def has_ub(self):
        """Returns :const:`False` when the upper bound is
        :const:`None` or positive infinity"""
        ub = self.upper
        return (ub is not None) and (ub != float('inf'))

    def lslack(self):
        """Lower slack (body - lb). Returns :const:`None` if
        a value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        lb = self.lower
        if lb is None:
            lb = -float('inf')
        else:
            lb = value(lb)
        return body - lb

    def uslack(self):
        """Upper slack (ub - body). Returns :const:`None` if
        a value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        ub = self.upper
        if ub is None:
            ub = float('inf')
        else:
            ub = value(ub)
        return ub - body

    def slack(self):
        """min(lslack, uslack). Returns :const:`None` if a
        value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        return min(self.lslack(), self.uslack())

    #
    # Override some default implementations on ComponentData
    #

    def index(self):
        return self._index

    #
    # Abstract Interface (ConstraintData)
    #

    @property
    def body(self):
        """Access the body of a constraint expression."""
        comp = self.parent_component()
        index = self._index
        data = comp._A_data
        indices = comp._A_indices
        indptr = comp._A_indptr
        x = comp._x
        ptrs = range(indptr[index], indptr[index + 1])
        return LinearExpression(
            linear_vars=[x[indices[p]] for p in ptrs],
            linear_coefs=[data[p] for p in ptrs],
            constant=0,
        )

    @property
    def lower(self):
        """Access the lower bound of a constraint
        expression."""
        comp = self.parent_component()
        index = self._index
        return comp._lower[index]

    @property
    def upper(self):
        """Access the upper bound of a constraint
        expression."""
        comp = self.parent_component()
        index = self._index
        return comp._upper[index]

    @property
    def equality(self):
        """A boolean indicating whether this is an equality
        constraint."""
        comp = self.parent_component()
        index = self._index
        if (comp._lower[index] is None) or (comp._upper[index] is None):
            return False
        return comp._lower[index] == comp._upper[index]

    @property
    def strict_lower(self):
        """A boolean indicating whether this constraint has
        a strict lower bound."""
        return False

    @property
    def strict_upper(self):
        """A boolean indicating whether this constraint has
        a strict upper bound."""
        return False

    def set_value(self, expr):
        """Set the expression on this constraint."""
        raise NotImplementedError("MatrixConstraint row elements can not be updated")


@ModelComponentFactory.register("A set of constraint expressions in Ax=b form.")
class MatrixConstraint(Mapping, IndexedConstraint):
    """
    Defines a set of linear constraints of the form:

       lb <= Ax <= ub

    where A is specified in the standard compressed sparse
    row (CSR) format. Variables must be provided as a list,
    whose ordering maps the variables to their column index
    in the associated coefficient matrix. This modeling
    component allows for fast construction of large linear
    constraint sets as it bypasses Pyomo's expression
    system.

    Parameters
    ----------
    A_data : list
        The values of the CSR format sparse matrix
    A_indices : list
        The column indices of the CSR format sparse matrix
    A_indptr : list
        The row start-stop pointers of the CSR format sparse matrix
    lb : list
        The list of constraint lower bounds
    ub : list
        The list of constraint upper bounds
    x : list
        The list of pyomo variables mapped to their appropriate column

    Example
    -------
    >>> from pyomo.environ import *
    >>> from pyomo.core.base.matrix_constraint import MatrixConstraint
    >>> model = ConcreteModel()
    >>>
    >>> # x_{i} <= x_{i+1}   (for i in {1,2})
    >>> model.v = Var(RangeSet(0,2))
    >>> data    = [1.0, -1.0, 1.0, -1.0]
    >>> indices = [  0,    1,   1,    2]
    >>> indptr  = [0, 2, 4]
    >>> lb      = [None, None]
    >>> ub      = [ 0.0,  0.0]
    >>> x       = [model.v[0], model.v[1], model.v[2]]
    >>> model.c = MatrixConstraint(data, indices, indptr, lb, ub, x)
    """

    def __init__(self, A_data, A_indices, A_indptr, lb, ub, x):
        m = len(lb)
        n = len(x)
        nnz = len(A_data)
        assert len(A_indices) == nnz
        assert len(A_indptr) == m + 1
        assert len(ub) == m
        IndexedConstraint.__init__(self, Any)

        self._A_data = A_data
        self._A_indices = A_indices
        self._A_indptr = A_indptr
        self._lower = lb
        self._upper = ub
        self._x = tuple(x)

    def construct(self, data=None):
        """Construct the expression(s) for this constraint."""
        if is_debug_set(logger):
            logger.debug("Constructing constraint %s" % (self.name))
        if self._constructed:
            return
        self._constructed = True

        ref = weakref.ref(self)
        with PauseGC():
            self._data = tuple(
                _MatrixConstraintData(i, ref) for i in range(len(self._lower))
            )

    #
    # Override some IndexedComponent methods
    #

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        return iter(i for i in range(len(self)))

    #
    # Remove methods that allow modifying this constraint
    #

    def add(self, index, expr):  # pragma:nocover
        raise NotImplementedError

    def __delitem__(self):  # pragma:nocover
        raise NotImplementedError

    def __setitem__(self, key, value):  # pragma:nocover
        raise NotImplementedError



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.core.base import ConcreteModel, Var, Reals
from pyomo.core.beta.list_objects import (
    XVarList,
    XConstraintList,
    XObjectiveList,
    XExpressionList,
)
from pyomo.core.base.var import VarData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.base.expression import ExpressionData


class _TestComponentListBase:
    _ctype = None
    _cdatatype = None

    def setUp(self):
        self.model = ConcreteModel()
        self.model.x = Var()
        # set by derived class
        self._arg = None

    def tearDown(self):
        self.model = None
        self._arg = None

    def test_init1(self):
        model = self.model
        model.c = self._ctype()
        self.assertEqual(model.c.is_indexed(), True)
        self.assertEqual(model.c.is_constructed(), True)

    def test_init2(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(model.c.is_indexed(), True)
        self.assertEqual(model.c.is_constructed(), True)
        with self.assertRaises(TypeError):
            model.d = self._ctype(*tuple(self._cdatatype(self._arg()) for i in index))

    def test_len1(self):
        model = self.model
        model.c = self._ctype()
        self.assertEqual(len(model.c), 0)

    def test_len2(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))

    def test_append(self):
        model = self.model
        model.c = self._ctype()
        index = range(5)
        self.assertEqual(len(model.c), 0)
        for i in index:
            c_new = self._cdatatype(self._arg())
            model.c.append(c_new)
            self.assertEqual(id(model.c[-1]), id(c_new))
            self.assertEqual(len(model.c), i + 1)

    def test_insert(self):
        model = self.model
        model.c = self._ctype()
        index = range(5)
        self.assertEqual(len(model.c), 0)
        for i in index:
            c_new = self._cdatatype(self._arg())
            model.c.insert(0, c_new)
            self.assertEqual(id(model.c[0]), id(c_new))
            self.assertEqual(len(model.c), i + 1)

    def test_setitem(self):
        model = self.model
        model.c = self._ctype()
        index = range(5)
        for i in index:
            model.c.append(self._cdatatype(self._arg()))
        for i in index:
            c_new = self._cdatatype(self._arg())
            self.assertNotEqual(id(c_new), id(model.c[i]))
            model.c[i] = c_new
            self.assertEqual(len(model.c), len(index))
            self.assertEqual(id(c_new), id(model.c[i]))

    def test_wrong_type_init(self):
        model = self.model
        index = range(5)
        with self.assertRaises(TypeError):
            model.c = self._ctype(self._arg() for i in index)

    def test_wrong_type_append(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        with self.assertRaises(TypeError):
            model.c.append(self._arg())

    def test_wrong_type_insert(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        model.c.insert(0, self._cdatatype(self._arg()))
        with self.assertRaises(TypeError):
            model.c.insert(0, self._arg())

    def test_wrong_type_setitem(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        model.c[0] = self._cdatatype(self._arg())
        with self.assertRaises(TypeError):
            model.c[0] = self._arg()

    def test_has_parent_init(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        with self.assertRaises(ValueError):
            model.c.append(model.c[0])
        with self.assertRaises(ValueError):
            model.d = self._ctype(model.c)

    def test_has_parent_append(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        with self.assertRaises(ValueError):
            model.c.append(model.c[0])
        d = []
        d.append(model.c[0])
        model.d = self._ctype()
        with self.assertRaises(ValueError):
            model.d.append(model.c[0])

    def test_has_parent_insert(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        model.c.insert(0, self._cdatatype(self._arg()))
        with self.assertRaises(ValueError):
            model.c.insert(0, model.c[0])
        d = []
        d.insert(0, model.c[0])
        model.d = self._ctype()
        with self.assertRaises(ValueError):
            model.d.insert(0, model.c[0])

    def test_has_parent_setitem(self):
        model = self.model
        model.c = self._ctype()
        model.c.append(self._cdatatype(self._arg()))
        model.c[0] = self._cdatatype(self._arg())
        model.c[0] = model.c[0]
        model.c.append(self._cdatatype(self._arg()))
        with self.assertRaises(ValueError):
            model.c[0] = model.c[1]

    # make sure an existing Data object IS replaced
    # by a call to setitem and not simply updated.
    def test_setitem_exists_overwrite(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        for i in index:
            cdata = model.c[i]
            self.assertEqual(id(cdata.parent_component()), id(model.c))
            model.c[i] = self._cdatatype(self._arg())
            self.assertEqual(len(model.c), len(index))
            self.assertNotEqual(id(cdata), id(model.c[i]))
            self.assertEqual(cdata.parent_component(), None)

    def test_delitem(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        for i in index:
            cdata = model.c[0]
            self.assertEqual(id(cdata.parent_component()), id(model.c))
            del model.c[0]
            self.assertEqual(len(model.c), len(index) - (i + 1))
            self.assertEqual(cdata.parent_component(), None)

    def test_iter(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        raw_list = model.c[:]
        self.assertEqual(type(raw_list), list)
        for c1, c2 in zip(raw_list, model.c):
            self.assertEqual(id(c1), id(c2))

    def test_reverse(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        raw_list = model.c[:]
        self.assertEqual(type(raw_list), list)
        model.c.reverse()
        raw_list.reverse()
        for c1, c2 in zip(model.c, raw_list):
            self.assertEqual(id(c1), id(c2))

    def test_remove(self):
        model = self.model
        model = ConcreteModel()
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        for i in index:
            cdata = model.c[0]
            self.assertEqual(cdata in model.c, True)
            model.c.remove(cdata)
            self.assertEqual(cdata in model.c, False)

    def test_pop(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        for i in index:
            cdata = model.c[-1]
            self.assertEqual(cdata in model.c, True)
            last = model.c.pop()
            self.assertEqual(cdata in model.c, False)
            self.assertEqual(id(cdata), id(last))

    def test_index(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        for i in index:
            cdata = model.c[i]
            self.assertEqual(model.c.index(cdata), i)
            self.assertEqual(model.c.index(cdata, start=i), i)
            with self.assertRaises(ValueError):
                model.c.index(cdata, start=i + 1)
            with self.assertRaises(ValueError):
                model.c.index(cdata, start=i, stop=i)
            with self.assertRaises(ValueError):
                model.c.index(cdata, stop=i)
            self.assertEqual(model.c.index(cdata, start=i, stop=i + 1), i)
            with self.assertRaises(ValueError):
                model.c.index(cdata, start=i + 1, stop=i + 1)
            self.assertEqual(model.c.index(cdata, start=-len(index) + i), i)
            if i == index[-1]:
                self.assertEqual(model.c.index(cdata, start=-len(index) + i + 1), i)
            else:
                with self.assertRaises(ValueError):
                    self.assertEqual(model.c.index(cdata, start=-len(index) + i + 1), i)
            if i == index[-1]:
                with self.assertRaises(ValueError):
                    self.assertEqual(model.c.index(cdata, stop=-len(index) + i + 1), i)
            else:
                self.assertEqual(model.c.index(cdata, stop=-len(index) + i + 1), i)
        tmp = self._cdatatype(self._arg())
        with self.assertRaises(ValueError):
            model.c.index(tmp)
        with self.assertRaises(ValueError):
            model.c.index(tmp, stop=len(model.c) + 1)

    def test_extend(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        c_more_list = [self._cdatatype(self._arg()) for i in index]
        self.assertEqual(len(model.c), len(index))
        self.assertTrue(len(c_more_list) > 0)
        for cdata in c_more_list:
            self.assertEqual(cdata.parent_component(), None)
        model.c.extend(c_more_list)
        for cdata in c_more_list:
            self.assertEqual(id(cdata.parent_component()), id(model.c))

    def test_count(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        for i in index:
            self.assertEqual(model.c.count(model.c[i]), 1)

    def test_model_clone(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        inst = model.clone()
        self.assertNotEqual(id(inst.c), id(model.c))
        for i in index:
            self.assertNotEqual(id(inst.c[i]), id(model.c[i]))

    def test_name(self):
        model = self.model
        index = range(5)
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        prefix = "c"
        for i in index:
            cdata = model.c[i]
            self.assertEqual(cdata.local_name, cdata.name)
            cname = prefix + "[" + str(i) + "]"
            self.assertEqual(cdata.local_name, cname)


class _TestActiveComponentListBase(_TestComponentListBase):
    def test_activate(self):
        model = self.model
        index = list(range(4))
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        self.assertEqual(model.c.active, True)
        model.c._active = False
        for i in index:
            model.c[i]._active = False
        self.assertEqual(model.c.active, False)
        for i in index:
            self.assertEqual(model.c[i].active, False)
        model.c.activate()
        self.assertEqual(model.c.active, True)

    def test_activate(self):
        model = self.model
        index = list(range(4))
        model.c = self._ctype(self._cdatatype(self._arg()) for i in index)
        self.assertEqual(len(model.c), len(index))
        self.assertEqual(model.c.active, True)
        for i in index:
            self.assertEqual(model.c[i].active, True)
        model.c.deactivate()
        self.assertEqual(model.c.active, False)
        for i in index:
            self.assertEqual(model.c[i].active, False)

    # Befault, assigning a new component to a dict container makes it
    # active (unless the default active state for the container
    # datatype is False, which is not the case for any currently
    # existing implementations).
    def test_active(self):
        model = self.model
        model = ConcreteModel()
        model.c = self._ctype()
        self.assertEqual(model.c.active, True)
        model.c.deactivate()
        self.assertEqual(model.c.active, False)
        model.c.append(self._cdatatype(self._arg()))
        self.assertEqual(model.c.active, True)
        model.c.deactivate()
        self.assertEqual(model.c.active, False)
        model.c.insert(0, self._cdatatype(self._arg()))
        self.assertEqual(model.c.active, True)


class TestVarList(_TestComponentListBase, unittest.TestCase):
    # Note: the updated VarData class only takes an optional
    # parent argument (you no longer pass the domain in)
    _ctype = XVarList
    _cdatatype = lambda self, arg: VarData()

    def setUp(self):
        _TestComponentListBase.setUp(self)
        self._arg = lambda: Reals


class TestExpressionList(_TestComponentListBase, unittest.TestCase):
    _ctype = XExpressionList
    _cdatatype = ExpressionData

    def setUp(self):
        _TestComponentListBase.setUp(self)
        self._arg = lambda: self.model.x**3


#
# Test components that include activate/deactivate
# functionality.
#


class TestConstraintList(_TestActiveComponentListBase, unittest.TestCase):
    _ctype = XConstraintList
    _cdatatype = ConstraintData

    def setUp(self):
        _TestComponentListBase.setUp(self)
        self._arg = lambda: self.model.x >= 1


class TestObjectiveList(_TestActiveComponentListBase, unittest.TestCase):
    _ctype = XObjectiveList
    _cdatatype = ObjectiveData

    def setUp(self):
        _TestComponentListBase.setUp(self)
        self._arg = lambda: self.model.x**2


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import sys

from pyomo.common.collections import Bunch
from pyomo.opt import ProblemFormat
from pyomo.core.base import Objective, Var, Constraint, value, ConcreteModel

_format = None


def convert(options=Bunch(), parser=None, model_format=None):
    global _format
    if not model_format is None:
        _format = model_format
    #
    # Import plugins
    #
    import pyomo.environ

    if options.model.save_file is None:
        if _format == ProblemFormat.cpxlp:
            options.model.save_file = 'unknown.lp'
        else:
            options.model.save_file = 'unknown.' + str(_format)
    options.model.save_format = _format

    data = Bunch(options=options)

    model_data = None
    try:
        pyomo.scripting.util.setup_environment(data)

        pyomo.scripting.util.apply_preprocessing(data, parser=parser)

        if data.error:
            return Bunch()

        model_data = pyomo.scripting.util.create_model(data)

        model_data.options = options
    except:
        # TBD: I should be able to call this function in the case of
        #      an exception to perform cleanup. However, as it stands
        #      calling finalize with its default keyword value for
        #      model(=None) results in an a different error related to
        #      task port values.  Not sure how to interpret that.
        pyomo.scripting.util.finalize(
            data, model=ConcreteModel(), instance=None, results=None
        )
        raise

    else:
        pyomo.scripting.util.finalize(data, model=model_data.model)

    return model_data


def convert_dakota(options=Bunch(), parser=None):
    #
    # Import plugins
    #
    import pyomo.environ

    model_file = os.path.basename(options.model.save_file)
    model_file_no_ext = os.path.splitext(model_file)[0]

    #
    # Set options for writing the .nl and related files
    #

    # By default replace .py with .nl
    if options.model.save_file is None:
        options.model.save_file = model_file_no_ext + '.nl'
    options.model.save_format = ProblemFormat.nl
    # Dakota requires .row/.col files
    options.model.symbolic_solver_labels = True

    #
    # Call the core converter
    #
    model_data = convert(options, parser)

    #
    # Generate Dakota input file fragments for the Vars, Objectives, Constraints
    #

    # TODO: the converted model doesn't expose the right symbol_map
    #       for only the vars active in the .nl

    model = model_data.instance

    # Easy way
    # print "VARIABLE:"
    # lines = open(options.save_model.replace('.nl','.col'),'r').readlines()
    # for varName in lines:
    #    varName = varName.strip()
    #    var = model_data.symbol_map.getObject(varName)
    #    print "'%s': %s" % (varName, var)
    #    #print var.pprint()

    # Hard way
    variables = 0
    var_descriptors = []
    var_lb = []
    var_ub = []
    var_initial = []
    tmpDict = model_data.symbol_map.getByObjectDictionary()
    for var in model.component_data_objects(Var, active=True):
        if id(var) in tmpDict:
            variables += 1
            var_descriptors.append(var.name)

            # apply user bound, domain bound, or infinite
            _lb, _ub = var.bounds
            if _lb is not None:
                var_lb.append(str(_lb))
            else:
                var_lb.append("-inf")

            if _ub is not None:
                var_ub.append(str(_ub))
            else:
                var_ub.append("inf")

            try:
                val = value(var)
            except:
                val = None
            var_initial.append(str(val))

    objectives = 0
    obj_descriptors = []
    for obj in model.component_data_objects(Objective, active=True):
        objectives += 1
        obj_descriptors.append(obj.name)

    constraints = 0
    cons_descriptors = []
    cons_lb = []
    cons_ub = []
    for con in model.component_data_objects(Constraint, active=True):
        constraints += 1
        cons_descriptors.append(con.name)
        if con.lower is not None:
            cons_lb.append(str(con.lower))
        else:
            cons_lb.append("-inf")
        if con.upper is not None:
            cons_ub.append(str(con.upper))
        else:
            cons_ub.append("inf")

    # Write the Dakota input file fragments

    dakfrag = open(model_file_no_ext + ".dak", 'w')

    dakfrag.write("#--- Dakota variables block ---#\n")
    dakfrag.write("variables\n")
    dakfrag.write("  continuous_design " + str(variables) + '\n')
    dakfrag.write("    descriptors\n")
    for vd in var_descriptors:
        dakfrag.write("      '%s'\n" % vd)
    dakfrag.write("    lower_bounds " + " ".join(var_lb) + '\n')
    dakfrag.write("    upper_bounds " + " ".join(var_ub) + '\n')
    dakfrag.write("    initial_point " + " ".join(var_initial) + '\n')

    dakfrag.write("#--- Dakota interface block ---#\n")
    dakfrag.write("interface\n")
    dakfrag.write("  algebraic_mappings = '" + options.model.save_file + "'\n")

    dakfrag.write("#--- Dakota responses block ---#\n")
    dakfrag.write("responses\n")
    dakfrag.write("  objective_functions " + str(objectives) + '\n')

    if constraints > 0:
        dakfrag.write("  nonlinear_inequality_constraints " + str(constraints) + '\n')
        dakfrag.write("    lower_bounds " + " ".join(cons_lb) + '\n')
        dakfrag.write("    upper_bounds " + " ".join(cons_ub) + '\n')

    dakfrag.write("    descriptors\n")
    for od in obj_descriptors:
        dakfrag.write("      '%s'\n" % od)
    if constraints > 0:
        for cd in cons_descriptors:
            dakfrag.write("      '%s'\n" % cd)

    # TODO: detect whether gradient information available in model
    dakfrag.write("  analytic_gradients\n")
    dakfrag.write("  no_hessians\n")

    dakfrag.close()

    sys.stdout.write(
        "Dakota input fragment written to file '%s'\n" % (model_file_no_ext + ".dak",)
    )
    return model_data


def pyomo2lp(args=None):
    from pyomo.scripting.pyomo_main import main

    if args is None:
        return main()
    else:
        return main(['convert', '--format=lp'] + args)


def pyomo2nl(args=None):
    from pyomo.scripting.pyomo_main import main

    if args is None:
        return main()
    else:
        return main(['convert', '--format=nl'] + args)


def pyomo2bar(args=None):
    from pyomo.scripting.pyomo_main import main

    if args is None:
        return main()
    else:
        return main(['convert', '--format=bar'] + args)


def pyomo2dakota(args=None):
    from pyomo.scripting.pyomo_main import main

    if args is None:
        return main()
    else:
        return main(['convert', '--format=dakota'] + args)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.interfaces.nlp import NLP, ExtendedNLP
import numpy as np
import scipy.sparse as sp


class _BaseNLPDelegator(NLP):
    def __init__(self, original_nlp):
        """
        This is a base class to make it easier to implement NLP
        classes that wrap other NLP instances. This class
        simply reproduces the NLP interface by passing the call
        onto the original_nlp passed in the constructor. This allows
        new wrapper classes to only implement the methods that change,
        allowing the others to pass through.

        Parameters
        ----------
        original_nlp : NLP-like
            The original NLP object that we want to wrap
        """
        super(NLP, self).__init__()
        self._original_nlp = original_nlp

    def n_primals(self):
        return self._original_nlp.n_primals()

    def primals_names(self):
        return self._original_nlp.primals_names()

    def n_constraints(self):
        return self._original_nlp.n_constraints()

    def constraint_names(self):
        return self._original_nlp.constraint_names()

    def nnz_jacobian(self):
        return self._original_nlp.nnz_jacobian()

    def nnz_hessian_lag(self):
        return self._original_nlp.nnz_hessian_lag()

    def primals_lb(self):
        return self._original_nlp.primals_lb()

    def primals_ub(self):
        return self._original_nlp.primals_ub()

    def constraints_lb(self):
        return self._original_nlp.constraints_lb()

    def constraints_ub(self):
        return self._original_nlp.constraints_ub()

    def init_primals(self):
        return self._original_nlp.init_primals()

    def init_duals(self):
        return self._original_nlp.init_duals()

    def create_new_vector(self, vector_type):
        return self._original_nlp.create_new_vector(vector_type)

    def set_primals(self, primals):
        self._original_nlp.set_primals(primals)

    def get_primals(self):
        return self._original_nlp.get_primals()

    def set_duals(self, duals):
        self._original_nlp.set_duals(duals)

    def get_duals(self):
        return self._original_nlp.get_duals()

    def set_obj_factor(self, obj_factor):
        self._original_nlp.set_obj_factor(obj_factor)

    def get_obj_factor(self):
        return self._original_nlp.get_obj_factor()

    def get_obj_scaling(self):
        return self._original_nlp.get_obj_scaling()

    def get_primals_scaling(self):
        return self._original_nlp.get_primals_scaling()

    def get_constraints_scaling(self):
        return self._original_nlp.get_constraints_scaling()

    def evaluate_objective(self):
        return self._original_nlp.evaluate_objective()

    def evaluate_grad_objective(self, out=None):
        return self._original_nlp.evaluate_grad_objective(out)

    def evaluate_constraints(self, out=None):
        return self._original_nlp.evaluate_constraints(out)

    def evaluate_jacobian(self, out=None):
        return self._original_nlp.evaluate_jacobian(out)

    def evaluate_hessian_lag(self, out=None):
        return self._original_nlp.evaluate_hessian_lag(out)

    def report_solver_status(self, status_code, status_message):
        self._original_nlp.report_solver_status(status_code, status_message)


class _ExtendedNLPDelegator(_BaseNLPDelegator):
    def __init__(self, original_nlp):
        if not isinstance(original_nlp, ExtendedNLP):
            raise TypeError(
                "Original NLP must be an instance of ExtendedNLP to use in"
                " an _ExtendedNLPDelegator. Got type %s" % type(original_nlp)
            )
        super().__init__(original_nlp)

    def n_eq_constraints(self):
        return self._original_nlp.n_eq_constraints()

    def n_ineq_constraints(self):
        return self._original_nlp.n_ineq_constraints()

    def evaluate_eq_constraints(self):
        return self._original_nlp.evaluate_eq_constraints()

    def evaluate_jacobian_eq(self):
        return self._original_nlp.evaluate_jacobian_eq()

    def evaluate_ineq_constraints(self):
        return self._original_nlp.evaluate_ineq_constraints()

    def evaluate_jacobian_ineq(self):
        return self._original_nlp.evaluate_jacobian_ineq()


class RenamedNLP(_BaseNLPDelegator):
    def __init__(self, original_nlp, primals_name_map):
        """
        This class takes an NLP that and allows one to rename the primals.
        It is a thin wrapper around the original NLP.

        Parameters
        ----------
        original_nlp : NLP-like
            The original NLP object that implements the NLP interface

        primals_name_map : dict of str --> str
            This is a dictionary that maps from the names
            in the original NLP class to the desired names
            for this instance.
        """
        super(RenamedNLP, self).__init__(original_nlp)
        self._primals_name_map = primals_name_map
        self._new_primals_names = None
        # Todo: maybe do this on first call instead of __init__?
        self._generate_new_names()

    def _generate_new_names(self):
        if self._new_primals_names is None:
            assert self._original_nlp.n_primals() == len(self._primals_name_map)
            self._new_primals_names = [
                self._primals_name_map[nm] for nm in self._original_nlp.primals_names()
            ]

    def primals_names(self):
        return self._new_primals_names


class ProjectedNLP(_BaseNLPDelegator):
    def __init__(self, original_nlp, primals_ordering):
        """
        This class takes an NLP that depends on a set of primals (original
        space) and converts it to an NLP that depends on a reordered set of
        primals (projected space).

        This will impact all the returned items associated with primal
        variables. E.g., the gradient will be in the new primals ordering
        instead of the original primals ordering.

        Note also that this can include additional primal variables not
        in the original NLP, or can exclude primal variables that were
        in the original NLP.

        Parameters
        ----------
        original_nlp : NLP-like
            The original NLP object that implements the NLP interface

        primals_ordering: list
           List of strings indicating the desired primal variable
           ordering for this NLP. The list can contain new variables
           that are not in the original NLP, thereby expanding the
           space of the primal variables.
        """
        super(ProjectedNLP, self).__init__(original_nlp)
        self._primals_ordering = list(primals_ordering)
        self._original_idxs = None
        self._projected_idxs = None
        self._generate_maps()
        self._projected_primals = self.init_primals()
        self._jacobian_nz_mask = None
        self._hessian_nz_mask = None
        self._nnz_jacobian = None
        self._nnz_hessian_lag = None

    def _generate_maps(self):
        if self._original_idxs is None or self._projected_idxs is None:
            primals_ordering_dict = {k: i for i, k in enumerate(self._primals_ordering)}
            original_names = self._original_nlp.primals_names()
            original_idxs = list()
            projected_idxs = list()
            for i, nm in enumerate(original_names):
                if nm in primals_ordering_dict:
                    # we need the reordering for this element
                    original_idxs.append(i)
                    projected_idxs.append(primals_ordering_dict[nm])
            self._original_idxs = np.asarray(original_idxs)
            self._projected_idxs = np.asarray(projected_idxs)
            self._original_to_projected = np.nan * np.zeros(
                self._original_nlp.n_primals()
            )
            self._original_to_projected[self._original_idxs] = self._projected_idxs

    def n_primals(self):
        return len(self._primals_ordering)

    def primals_names(self):
        return list(self._primals_ordering)

    def nnz_jacobian(self):
        if self._nnz_jacobian is None:
            J = self.evaluate_jacobian()
            self._nnz_jacobian = len(J.data)
        return self._nnz_jacobian

    def nnz_hessian_lag(self):
        if self._nnz_hessian_lag is None:
            H = self.evaluate_hessian_lag()
            self._nnz_hessian_lag = len(H.data)
        return self._nnz_hessian_lag

    def _project_primals(self, default, original_primals):
        projected_x = default * np.ones(self.n_primals(), dtype=np.float64)
        projected_x[self._projected_idxs] = original_primals[self._original_idxs]
        return projected_x

    def primals_lb(self):
        return self._project_primals(-np.inf, self._original_nlp.primals_lb())

    def primals_ub(self):
        return self._project_primals(np.inf, self._original_nlp.primals_ub())

    def init_primals(self):
        # Todo: think about what to do here if an entry is not defined
        # for now, we default to NaN, but there may be a better way
        # (e.g., taking a new initial value in the constructor?)
        return self._project_primals(np.nan, self._original_nlp.init_primals())

    def create_new_vector(self, vector_type):
        if vector_type == 'primals':
            return np.zeros(self.n_primals(), dtype=np.float64)
        return self._original_nlp.create_new_vector(vector_type)

    def set_primals(self, primals):
        # here, we keep a local copy of the projected primals so
        # we can give back these values in get_primals. This might
        # not be the best idea since we can't really support this
        # same strategy for other methods (e.g., init_primals) where
        # we now use NaNs to fill in any "missing" entries.
        np.copyto(self._projected_primals, primals)
        original_primals = self._original_nlp.get_primals()
        original_primals[self._original_idxs] = primals[self._projected_idxs]
        self._original_nlp.set_primals(original_primals)

    def get_primals(self):
        original_primals = self._original_nlp.get_primals()
        self._projected_primals[self._projected_idxs] = original_primals[
            self._original_idxs
        ]
        return self._projected_primals

    def get_primals_scaling(self):
        return self._project_primals(np.nan, self._original_nlp.get_primals_scaling())

    def evaluate_grad_objective(self, out=None):
        original_grad_objective = self._original_nlp.evaluate_grad_objective()
        projected_objective = self._project_primals(0.0, original_grad_objective)
        if out is None:
            return projected_objective
        np.copyto(out, projected_objective)
        return out

    def evaluate_jacobian(self, out=None):
        original_jacobian = self._original_nlp.evaluate_jacobian()
        if out is not None:
            np.copyto(out.data, original_jacobian.data[self._jacobian_nz_mask])
            return out

        row = original_jacobian.row
        col = original_jacobian.col
        data = original_jacobian.data

        if self._jacobian_nz_mask is None:
            # need to remap the irow, jcol to the new space and change the size
            self._jacobian_nz_mask = np.isin(col, self._original_idxs)

        new_col = col[self._jacobian_nz_mask]
        new_col = self._original_to_projected[new_col]
        new_row = row[self._jacobian_nz_mask]
        new_data = data[self._jacobian_nz_mask]

        return sp.coo_matrix(
            (new_data, (new_row, new_col)),
            shape=(self.n_constraints(), self.n_primals()),
        )

    def evaluate_hessian_lag(self, out=None):
        original_hessian = self._original_nlp.evaluate_hessian_lag()
        if out is not None:
            np.copyto(out.data, original_hessian.data[self._hessian_nz_mask])
            return out

        row = original_hessian.row
        col = original_hessian.col
        data = original_hessian.data

        if self._hessian_nz_mask is None:
            # need to remap the irow, jcol to the new space and change the size
            self._hessian_nz_mask = np.isin(col, self._original_idxs) & np.isin(
                row, self._original_idxs
            )

        new_col = col[self._hessian_nz_mask]
        new_col = self._original_to_projected[new_col]
        new_row = row[self._hessian_nz_mask]
        new_row = self._original_to_projected[new_row]
        new_data = data[self._hessian_nz_mask]

        return sp.coo_matrix(
            (new_data, (new_row, new_col)), shape=(self.n_primals(), self.n_primals())
        )

    def report_solver_status(self, status_code, status_message):
        raise NotImplementedError('Need to think about this...')


class ProjectedExtendedNLP(ProjectedNLP, _ExtendedNLPDelegator):
    def __init__(self, original_nlp, primals_ordering):
        super(ProjectedExtendedNLP, self).__init__(original_nlp, primals_ordering)
        self._jacobian_eq_nz_mask = None
        self._jacobian_ineq_nz_mask = None

    def evaluate_jacobian_eq(self, out=None):
        original_jacobian = self._original_nlp.evaluate_jacobian_eq()
        if out is not None:
            np.copyto(out.data, original_jacobian.data[self._jacobian_eq_nz_mask])
            return out

        row = original_jacobian.row
        col = original_jacobian.col
        data = original_jacobian.data

        if self._jacobian_eq_nz_mask is None:
            # need to remap the irow, jcol to the new space and change the size
            self._jacobian_eq_nz_mask = np.isin(col, self._original_idxs)

        new_col = col[self._jacobian_eq_nz_mask]
        new_col = self._original_to_projected[new_col]
        new_row = row[self._jacobian_eq_nz_mask]
        new_data = data[self._jacobian_eq_nz_mask]

        return sp.coo_matrix(
            (new_data, (new_row, new_col)),
            shape=(self.n_eq_constraints(), self.n_primals()),
        )

    def evaluate_jacobian_ineq(self, out=None):
        original_jacobian = self._original_nlp.evaluate_jacobian_ineq()
        if out is not None:
            np.copyto(out.data, original_jacobian.data[self._jacobian_ineq_nz_mask])
            return out

        row = original_jacobian.row
        col = original_jacobian.col
        data = original_jacobian.data

        if self._jacobian_ineq_nz_mask is None:
            self._jacobian_ineq_nz_mask = np.isin(col, self._original_idxs)

        new_col = col[self._jacobian_ineq_nz_mask]
        new_col = self._original_to_projected[new_col]
        new_row = row[self._jacobian_ineq_nz_mask]
        new_data = data[self._jacobian_ineq_nz_mask]

        return sp.coo_matrix(
            (new_data, (new_row, new_col)),
            shape=(self.n_ineq_constraints(), self.n_primals()),
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io

import pyomo.environ as pyo
from pyomo.common import unittest
from pyomo.common.collections import ComponentMap
from pyomo.common.fileutils import this_file_dir
from pyomo.contrib.solver.solvers.asl_sol_reader import (
    ASLSolFileSolutionLoader,
    ASLSolFileData,
    asl_solve_code_to_solution_status,
    parse_asl_sol_file,
)
from pyomo.contrib.solver.common.results import (
    Results,
    SolutionStatus,
    TerminationCondition,
)
from pyomo.contrib.solver.common.util import SolverError
from pyomo.repn.plugins.nl_writer import NLWriterInfo, ScalingFactors


class TestSolFileData(unittest.TestCase):
    def test_default_instantiation(self):
        instance = ASLSolFileData()
        self.assertEqual(instance.message, None)
        self.assertEqual(instance.objno, 0)
        self.assertEqual(instance.solve_code, None)
        self.assertEqual(instance.ampl_options, None)
        self.assertEqual(instance.primals, None)
        self.assertEqual(instance.duals, None)
        self.assertEqual(instance.var_suffixes, {})
        self.assertEqual(instance.con_suffixes, {})
        self.assertEqual(instance.obj_suffixes, {})
        self.assertEqual(instance.problem_suffixes, {})
        self.assertEqual(instance.unparsed, None)


class TestSolParserUtils(unittest.TestCase):
    def test_solve_code_to_status(self):
        sol_data = ASLSolFileData()
        sol_data.message = None

        sol_data.solve_code = None
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.noSolution, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(None): solver did not generate a SOL file",
            r.extra_info.solver_message,
        )

        sol_data.primals = [1]
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(None): solver did not generate a SOL file",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 0
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.optimal, r.solution_status)
        self.assertEqual(
            TerminationCondition.convergenceCriteriaSatisfied, r.termination_condition
        )
        self.assertEqual("", r.extra_info.solver_message)

        sol_data.solve_code = 100
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.feasible, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(100:solved?): optimal solution indicated, but error likely",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 200
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.infeasible, r.solution_status)
        self.assertEqual(
            TerminationCondition.locallyInfeasible, r.termination_condition
        )
        self.assertEqual(
            "AMPL(200:infeasible): constraints cannot be satisfied",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 300
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.unbounded, r.termination_condition)
        self.assertEqual(
            "AMPL(300:unbounded): objective can be improved without limit",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 400
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.iterationLimit, r.termination_condition)
        self.assertEqual(
            "AMPL(400:limit): stopped by a limit that you set",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 500
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(500:failure): stopped by an error condition in the solver",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 99
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.optimal, r.solution_status)
        self.assertEqual(
            TerminationCondition.convergenceCriteriaSatisfied, r.termination_condition
        )
        self.assertEqual("", r.extra_info.solver_message)

        sol_data.solve_code = 199
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.feasible, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(199:solved?): optimal solution indicated, but error likely",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 299
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.infeasible, r.solution_status)
        self.assertEqual(
            TerminationCondition.locallyInfeasible, r.termination_condition
        )
        self.assertEqual(
            "AMPL(299:infeasible): constraints cannot be satisfied",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 399
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.unbounded, r.termination_condition)
        self.assertEqual(
            "AMPL(399:unbounded): objective can be improved without limit",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 499
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.iterationLimit, r.termination_condition)
        self.assertEqual(
            "AMPL(499:limit): stopped by a limit that you set",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = 599
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(599:failure): stopped by an error condition in the solver",
            r.extra_info.solver_message,
        )

        sol_data.solve_code = -1
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual("AMPL(-1): unexpected solve code", r.extra_info.solver_message)

        sol_data.solve_code = 600
        r = Results()
        asl_solve_code_to_solution_status(sol_data, r)
        self.assertEqual(SolutionStatus.unknown, r.solution_status)
        self.assertEqual(TerminationCondition.error, r.termination_condition)
        self.assertEqual(
            "AMPL(600): unexpected solve code", r.extra_info.solver_message
        )


class TestSolParser(unittest.TestCase):
    def test_parse_minimal_sol_file(self):
        # Build a tiny .sol text stream:
        # - "Options" block with number_of_options = 2, then 4 model_object ints
        #   model_objects[1] = #cons, model_objects[3] = #vars
        # - #cons duals lines
        # - #vars primals lines
        # - "objno <i> <exit_code>"
        n_cons = 2
        n_vars = 3
        stream = io.StringIO(f"""Solver message preamble
Options
2
1
2
{n_cons}
{n_cons}
{n_vars}
{n_vars}
1.5
-2.25
10.0
20.0
30.0
objno 0 100""")
        sol_data = parse_asl_sol_file(stream)

        self.assertEqual("Solver message preamble", sol_data.message)
        self.assertEqual(0, sol_data.objno)
        self.assertEqual(100, sol_data.solve_code)
        self.assertEqual([1, 2], sol_data.ampl_options)
        self.assertEqual([10.0, 20.0, 30.0], sol_data.primals)
        self.assertEqual([1.5, -2.25], sol_data.duals)
        self.assertEqual({}, sol_data.var_suffixes)
        self.assertEqual({}, sol_data.con_suffixes)
        self.assertEqual({}, sol_data.obj_suffixes)
        self.assertEqual({}, sol_data.problem_suffixes)
        self.assertEqual(None, sol_data.unparsed)

    def test_parse_vbtol(self):
        stream = io.StringIO(f"""Solver message preamble
Options
2
1
3
2
0
3
0
1.5
objno 0 100""")
        sol_data = parse_asl_sol_file(stream)

        self.assertEqual("Solver message preamble", sol_data.message)
        self.assertEqual(0, sol_data.objno)
        self.assertEqual(100, sol_data.solve_code)
        self.assertEqual([1, 3, 1.5], sol_data.ampl_options)
        self.assertEqual([], sol_data.primals)
        self.assertEqual([], sol_data.duals)
        self.assertEqual({}, sol_data.var_suffixes)
        self.assertEqual({}, sol_data.con_suffixes)
        self.assertEqual({}, sol_data.obj_suffixes)
        self.assertEqual({}, sol_data.problem_suffixes)
        self.assertEqual(None, sol_data.unparsed)

    def test_multiline_message_and_unparsed(self):
        stream = io.StringIO("""CONOPT 3.17A: Optimal; objective 1
4 iterations; evals: nf = 2, ng = 0, nc = 2, nJ = 0, nH = 0, nHv = 0

Options
3
1
1
0
1
1
1
1
1
1
objno 0 0
suffix 0 1 8 0 0
sstatus
0 1
suffix 1 1 8 0 0
sstatus
0 3
extra data here
and here
""")
        sol_data = parse_asl_sol_file(stream)

        self.assertEqual(
            "CONOPT 3.17A: Optimal; objective 1\n"
            "4 iterations; evals: nf = 2, ng = 0, nc = 2, nJ = 0, nH = 0, nHv = 0",
            sol_data.message,
        )
        self.assertEqual(0, sol_data.objno)
        self.assertEqual(0, sol_data.solve_code)
        self.assertEqual([1, 1, 0], sol_data.ampl_options)
        self.assertEqual([1.0], sol_data.primals)
        self.assertEqual([1.0], sol_data.duals)
        self.assertEqual({'sstatus': {0: 1}}, sol_data.var_suffixes)
        self.assertEqual({'sstatus': {0: 3}}, sol_data.con_suffixes)
        self.assertEqual({}, sol_data.obj_suffixes)
        self.assertEqual({}, sol_data.problem_suffixes)
        self.assertEqual("extra data here\nand here\n", sol_data.unparsed)

    def test_suffix_table(self):
        stream = io.StringIO("""CONOPT 3.17A: Optimal; objective 1
4 iterations; evals: nf = 2, ng = 0, nc = 2, nJ = 0, nH = 0, nHv = 0

Options
3
1
1
0
1
1
1
1
1
1
objno 0 0
suffix 0 1 7 36 3
custom
1 INT An int field
2 DBL double
3 STR
0 1
suffix 1 1 8 0 0
sstatus
0 3
suffix 2 1 8 0 0
sstatus
0 2
suffix 3 1 8 0 0
sstatus
0 4

""")
        sol_data = parse_asl_sol_file(stream)

        self.assertEqual(
            "CONOPT 3.17A: Optimal; objective 1\n"
            "4 iterations; evals: nf = 2, ng = 0, nc = 2, nJ = 0, nH = 0, nHv = 0",
            sol_data.message,
        )
        self.assertEqual(0, sol_data.objno)
        self.assertEqual(0, sol_data.solve_code)
        self.assertEqual([1, 1, 0], sol_data.ampl_options)
        self.assertEqual([1.0], sol_data.primals)
        self.assertEqual([1.0], sol_data.duals)
        self.assertEqual({'custom': {0: 1}}, sol_data.var_suffixes)
        self.assertEqual({'sstatus': {0: 3}}, sol_data.con_suffixes)
        self.assertEqual({'sstatus': {0: 2}}, sol_data.obj_suffixes)
        self.assertEqual({'sstatus': 4}, sol_data.problem_suffixes)
        self.assertEqual(
            {
                (0, 'custom'): [
                    [1, 'INT', 'An int field'],
                    [2, 'DBL', 'double'],
                    [3, 'STR'],
                ]
            },
            sol_data.suffix_table,
        )
        self.assertEqual(None, sol_data.unparsed)

    def test_error_missing_options(self):
        # No line contains the substring "Options"
        bad_text = "Solver message preamble\nNo header here\n"
        stream = io.StringIO(bad_text)

        with self.assertRaisesRegex(
            SolverError, "Error reading `sol` file: no 'Options' line found."
        ):
            parse_asl_sol_file(stream)

    def test_error_malformed_options(self):
        # Contains "Options" but the required integer line is missing/blank
        bad_text = "Preamble\nOptions\n\n"
        stream = io.StringIO(bad_text)

        with self.assertRaisesRegex(ValueError, "invalid literal"):
            parse_asl_sol_file(stream)

    def test_error_objno_not_found(self):
        stream = io.StringIO(f"""Solver message preamble
Options
2
1
2
2
0
3
0
1.5
objno 0""")

        with self.assertRaisesRegex(
            SolverError,
            "Error reading `sol` file: expected 'objno'; " "received '1.5'.",
        ):
            sol_data = parse_asl_sol_file(stream)

    def test_error_objno_bad_format(self):
        stream = io.StringIO(f"""Solver message preamble
Options
2
1
2
2
0
3
0
objno 0""")

        with self.assertRaisesRegex(
            SolverError,
            "Error reading `sol` file: expected two numbers in 'objno' line; "
            "received 'objno 0'.",
        ):
            sol_data = parse_asl_sol_file(stream)


class TestSolFileSolutionLoader(unittest.TestCase):

    def test_member_list(self):
        expected_list = ['load_vars', 'get_primals', 'get_duals', 'get_reduced_costs']
        method_list = [
            method
            for method in dir(ASLSolFileSolutionLoader)
            if not method.startswith('_')
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_load_vars(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])

        nl_info = NLWriterInfo(var=[m.x, m.y[1], m.y[3]])
        sol_data = ASLSolFileData()
        sol_data.primals = [3, 7, 5]
        loader = ASLSolFileSolutionLoader(sol_data, nl_info)

        loader.load_vars()
        self.assertEqual(m.x.value, 3)
        self.assertEqual(m.y[1].value, 7)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, 5)

        sol_data.primals = [13, 17, 15]
        loader.load_vars(vars_to_load=[m.y[3], m.x])
        self.assertEqual(m.x.value, 13)
        self.assertEqual(m.y[1].value, 7)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, 15)

        nl_info.scaling = ScalingFactors([1, 5, 10], [], [])
        loader.load_vars()
        self.assertEqual(m.x.value, 13)
        self.assertEqual(m.y[1].value, 3.4)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, 1.5)

        nl_info.eliminated_vars = [(m.y[2], 2 * m.y[3] + 1)]
        loader.load_vars()
        self.assertEqual(m.x.value, 13)
        self.assertEqual(m.y[1].value, 3.4)
        self.assertEqual(m.y[2].value, 4)
        self.assertEqual(m.y[3].value, 1.5)

    def test_load_vars_empty_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])

        nl_info = NLWriterInfo(
            var=[], eliminated_vars=[(m.y[3], 1.5), (m.y[2], 2 * m.y[3] + 1)]
        )
        sol_data = ASLSolFileData()
        sol_data.primals = []
        loader = ASLSolFileSolutionLoader(sol_data, nl_info)

        loader.load_vars()
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, 4)
        self.assertEqual(m.y[3].value, 1.5)

    def test_get_primals(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])

        nl_info = NLWriterInfo(var=[m.x, m.y[1], m.y[3]])
        sol_data = ASLSolFileData()
        sol_data.primals = [3, 7, 5]
        loader = ASLSolFileSolutionLoader(sol_data, nl_info)

        self.assertEqual(
            loader.get_primals(), ComponentMap([(m.x, 3), (m.y[1], 7), (m.y[3], 5)])
        )
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, None)

        sol_data.primals = [13, 17, 15]
        self.assertEqual(
            loader.get_primals(vars_to_load=[m.y[3], m.x]),
            ComponentMap([(m.x, 13), (m.y[3], 15)]),
        )
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, None)

        nl_info.scaling = ScalingFactors([1, 5, 10], [], [])
        self.assertEqual(
            loader.get_primals(),
            ComponentMap([(m.x, 13), (m.y[1], 3.4), (m.y[3], 1.5)]),
        )
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, None)

        nl_info.eliminated_vars = [(m.y[2], 2 * m.y[3] + 1)]
        self.assertEqual(
            loader.get_primals(),
            ComponentMap([(m.x, 13), (m.y[1], 3.4), (m.y[2], 4), (m.y[3], 1.5)]),
        )
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, None)

    def test_get_primals_empty_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])

        nl_info = NLWriterInfo(
            var=[], eliminated_vars=[(m.y[3], 1.5), (m.y[2], 2 * m.y[3] + 1)]
        )
        sol_data = ASLSolFileData()
        sol_data.primals = []
        loader = ASLSolFileSolutionLoader(sol_data, nl_info)

        self.assertEqual(
            loader.get_primals(), ComponentMap([(m.y[2], 4), (m.y[3], 1.5)])
        )
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y[1].value, None)
        self.assertEqual(m.y[2].value, None)
        self.assertEqual(m.y[3].value, None)
